' IDDEVICE.BAS, v1.0
'
' Copyright (C) 2002 Michael Webster
'
' This program is free software; you can redistribute it and/or
' modify it under the terms of the GNU General Public License as
' published by the Free Software Foundation; either version 2 of
' the License, or (at your option) any later version.
'
' This program is distributed in the hope that it will be useful,
' but WITHOUT ANY WARRANTY; without even the implied warranty of
' MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
' GNU General Public License for more details.
'
' You should have received a copy of the GNU General Public
' License along with this program; if not, write to the Free
' Software Foundation, Inc., 59 Temple Place, Suite 330,
' Boston, MA  02111-1307  USA

DECLARE SUB MainWin ()
DECLARE SUB InfoWin (listItems$(), device%)
DECLARE FUNCTION FirmwareRevision$ (device%)
DECLARE FUNCTION SerialNumber$ (device%)
DECLARE SUB MakeList (listItems$(), device%)
DECLARE FUNCTION ModelNumber$ (device%)
DECLARE SUB IdentifyDevice (device%, errorCode%)
DECLARE SUB ScanDevices (devices$())
DEFINT A-Z

' This file contains a QuickBASIC program that checks the standard ATA
' interface channels for devices that support the ATA or ATAPI Identify
' Device command, presents the user with a list of the devices that are
' present, and allows the user to select a device and view selected
' portions of the information returned by the device in response to the
' Identify Device command.
'
' IMPORTANT:
'   The IdentifyDevice procedure programs the device controller
'   directly and makes no (substantial) attempt to protect other
'   processes that may be using the device. For *best* reliability,
'   this program should be run under MS-DOS or Windows 9x MS-DOS
'   mode, without any disk cache or similar programs loaded.
'   If you modify the code to allow this program to run under
'   Windows, then you should take care to have no other programs
'   with unsaved data open. Refer to comments in the IdentifyDevice
'   procedure for more information.
'
' Windows 9x versions other than Windows 98 First Edition (4.10.1998)
' (effectively) block access to the Command/Control Block registers.
'
' To run this program within the QuickBASIC or QuickBASIC Extended
' programming environment, you must load the default Quick library
' (QB.QLB or QBX.QLB) when the programming environment is started.
' Refer to QUICKLIB.TXT for more information.
'
' For the current version of this program, the primary sources of
' information were the working drafts for various versions of the
' ATA and ATA/ATAPI standards (ATA/ATAPI-5 revision 3 being the
' most recent). For the current standards, try http://www.t13.org
'
' Although this program appears to work OK on the typical system
' with an integrated IDE interface, it may not work on some older
' systems that use an IDE interface adapter, or on systems with
' unusual configurations.
'
' The first version of this program was created because BIOSes were,
' at that time, unable to read the drive parameters from the new IDE
' drives and some drive makers were failing to print the parameters
' on their drives.
'
' UPDATE: In the initial distribution, this program obtained the
' device status by reading the device's Status Register. In an
' environment where other processes may be using the device, this
' could potentially cause a problem because reading the Status
' Register clears pending interrupts. This version of the program
' obtains the device status by reading the device's Alternate
' Status Register (which does not clear pending interrupts).

' $INCLUDE: 'INTX.BI'
' $INCLUDE: 'DIALOG.BI'
' $INCLUDE: 'GENERAL.BI'
' $INCLUDE: 'ERROR.BI'
' $INCLUDE: 'BIOSDISK.BI'
' $INCLUDE: 'MOUSE.BI'
' $INCLUDE: 'INP16.BI'

CONST cFalse = 0, cTrue = NOT cFalse

' Define constants for the base I/O addresses of the primary and
' secondary ATA interface channels.
CONST cIoBase1 = &H1F0
CONST cIoBase2 = &H170

' Define constants for the necessary Command/Control Block registers.
' The assigned values are the offset of the register from the base
' I/O address. Note that the Data Register, unlike all the other
' registers, is a 16-bit register accessed through a 16-bit I/O port.
CONST cRegData = 0          ' Data Register
CONST cSectorCount = 2      ' Sector Count Register
CONST cRegCylLow = 4        ' Cylinder Low Register
CONST cRegCylHigh = 5       ' Cylinder High Register
CONST cRegDeviceHead = 6    ' Device/Head Register
CONST cRegCommand = 7       ' Command Register (write)
CONST cRegStatus = 7        ' Status Register (read)
CONST cRegAltStatus = &H206 ' Alternate Status Register (read)

' Define constants for the Command/Control Block bit masks.
CONST cDeviceHeadDEV = &H10 ' Device (0 = master, 1 = slave)
CONST cStatusERR = 1        ' Error (last command returned an error)
CONST cStatusIDX = 2        ' Active once per revolution (not available
                            ' for all devices and/or interfaces)
CONST cStatusDRQ = 8        ' Data Request (device ready for transfer)
CONST cStatusDRDY = &H40    ' Device Ready (to accept a command)
CONST cStatusBSY = &H80     ' Device Busy

' Define constants for the Identify Device commands.
CONST cAtaIdentifyDevice = &HEC
CONST cAtapiIdentifyDevice = &HA1

' Define constants for the ATAPI signature. This signature is returned
' (in the cylinder high and cylinder low registers) by an ATAPI device
' in response to an ATA Identify Device command.
CONST cAtapiSignatureHigh = &HEB
CONST cAtapiSignatureLow = &H14

' Define constants for the error code values returned by
' the IdentifyDevice procedure.
CONST cIdErrorNoDevice = 1
CONST cIdErrorTimeoutBSY = 2
CONST cIdErrorTimeoutNoDRQ = 3
CONST cIdErrorUnexpected = 4

' Define constants for all of the devices that can be
' connected to the standard ATA interface channels.
CONST cPrimaryMaster = 1
CONST cPrimarySlave = 2
CONST cSecondaryMaster = 3
CONST cSecondarySlave = 4

' Declare a shared array to store the device information returned by the
' Identify Device command. The first dimension is indexed by the device.
DIM SHARED sDeviceInformation(1 TO 4, 256)

' Declare a shared array to store the detection state for each device.
DIM SHARED sfDeviceDetected(1 TO 4)

MainWin

END

FUNCTION FirmwareRevision$ (device%)

    ' Returns the 8 character firmware revision for the device specified
    ' by <device%> from the shared device information array.
  
    ' Return a null string if the firmware revision is not specified.
    IF sDeviceInformation(device, 23) = 0 THEN EXIT FUNCTION
  
    FOR i = 23 TO 26
        r$ = r$ + CHR$(HighByte(sDeviceInformation(device, i))) + CHR$(LowByte(sDeviceInformation(device, i)))
    NEXT
    FirmwareRevision$ = LTRIM$(r$)

END FUNCTION

SUB IdentifyDevice (device%, errorCode%)

    ' Reads 256 16-bit words of device information from the ATA or
    ' ATAPI device specified by <device%>. The information is read by
    ' sending the ATA or ATAPI Identify Device command to the device
    ' controller. If the device is present and the command is
    ' successful, then this procedure returns with <errorCode%> set
    ' to 0 and the device information copied to the elements of the
    ' shared device information array. Otherwise, it returns with
    ' <errorCode%> set to a value that indicates the type of error
    ' that occurred.
    '
    ' Constants for the <errorCode%> return value are defined in the
    ' main module.
    '
    ' IMPORTANT:
    '   This procedure contains the *minimum* code necessary for
    '   it to function with *reasonable* reliability. It accesses
    '   the Command/Control Block registers without performing the
    '   host and/or device status checks that are specified in the
    '   ATA standard. For example, it does not verify that the
    '   Device Busy and Data Request bits in the Alternate Status
    '   Register are clear before it writes to the Device/Head,
    '   Sector Count, and Command registers.

    ' To facilitate tuning of this procedure, declare a local constant
    ' for the delay period between register accesses.
    CONST cDelayMs = 10
    
    errorCode = 0

    ' Set local variables for all the necessary parameters.
    SELECT CASE device
        CASE cPrimaryMaster
            basePort = cIoBase1
        CASE cPrimarySlave
            basePort = cIoBase1
            fSlave = cTrue
        CASE cSecondaryMaster
            basePort = cIoBase2
        CASE cSecondarySlave
            basePort = cIoBase2
            fSlave = cTrue
    END SELECT
    regData = basePort + cRegData
    regSectorCount = basePort + cSectorCount
    regCylLow = basePort + cRegCylLow
    regCylHigh = basePort + cRegCylHigh
    regDeviceHead = basePort + cRegDeviceHead
    regCommand = basePort + cRegCommand
    regAltStatus = basePort + cRegAltStatus

    ' Save the value from the Device/Head Register.
    dhrValue = INP(regDeviceHead)

    Delay cDelayMs

    ' Set or clear the DEV bit (set for slave, clear for
    ' master) and write the value back to the register.
    IF fSlave THEN
        OUT regDeviceHead, dhrValue OR cDeviceHeadDEV
    ELSE
        OUT regDeviceHead, dhrValue AND NOT cDeviceHeadDEV
    END IF

    ' Determine if the Sector Count Register exists by writing
    ' a test value to the register and reading the value back.
    ' To create the test value, complement the current value
    ' and mask the upper byte (because the Sector Count register
    ' is an 8-bit register). If the register does not exist,
    ' indicating that no devices are connected to the channel,
    ' then exit and return the appropriate error code.
    sectorCount = INP(regSectorCount)
    testVal = NOT sectorCount AND &HFF
    Delay cDelayMs
    OUT regSectorCount, testVal
    Delay cDelayMs
    IF INP(regSectorCount) <> testVal THEN
        'PrintDebug "device=" + STR$(device)
        'PrintDebug "scr=" + HEX$(sectorCount)
        'PrintDebug "testval=" + HEX$(testVal)
        'PrintDebug "readback=" + HEX$(INP(regSectorCount))
        errorCode = cIdErrorNoDevice
        EXIT SUB
    END IF
    OUT regSectorCount, sectorCount

    ' Wait for the device to indicate that it is not busy. If the
    ' Device Busy bit in the Alternate Status Register is not clear
    ' within 10 delay periods, then exit and return the appropriate
    ' error code.
    FOR i = 1 TO 10
        Delay cDelayMs
        IF (INP(regAltStatus) AND cStatusBSY) = 0 THEN EXIT FOR
    NEXT
    IF i > 10 THEN
        errorCode = cIdErrorTimeoutBSY
        EXIT SUB
    END IF

    ' Send the ATA Identify Device command to the Command Register.
    ' For an ATAPI device, this command will return with the error
    ' bit in the Alternate Status Register set and the ATAPI
    ' signature in the cylinder high and cylinder low registers.
    OUT regCommand, cAtaIdentifyDevice
    Delay cDelayMs

    ' Check for the ATAPI signature and send the ATAPI
    ' Identify Device command if it is present.
    IF INP(regCylHigh) = cAtapiSignatureHigh AND INP(regCylLow) = cAtapiSignatureLow THEN
        
        ' Wait for the device to indicate that it is not busy.
        ' If the Device Busy bit in the Alternate Status Register
        ' is not clear within 10 delay periods, then exit and
        ' return the appropriate error code.
        FOR i = 1 TO 10
            Delay cDelayMs
            IF (INP(regAltStatus) AND cStatusBSY) = 0 THEN EXIT FOR
        NEXT
        IF i > 10 THEN
            errorCode = cIdErrorTimeoutBSY
            EXIT SUB
        END IF

        ' Send the ATAPI Identify Device command to the Command Register.
        OUT regCommand, cAtapiIdentifyDevice

    END IF

    ' Wait for the device to clear the Device Busy bit in the Alternate
    ' Status Register and set the Data Request bit. If the device does
    ' not do so within 5 delay periods, then exit and return the
    ' appropriate error code.
    FOR i = 1 TO 5
        Delay cDelayMs
        IF (INP(regAltStatus) AND (cStatusBSY OR cStatusDRQ)) = cStatusDRQ THEN
            EXIT FOR
        END IF
    NEXT
    IF i > 5 THEN
        errorCode = cIdErrorTimeoutNoDRQ
        EXIT SUB
    END IF

    ' If the last command returned an error, then
    ' exit and return the appropriate error code.
    IF INP(regAltStatus) AND cStatusERR THEN
        errorCode = cIdErrorUnexpected
        EXIT SUB
    END IF

    ' Read the contents of the device's sector buffer
    ' into the shared device information array.
    FOR i = 0 TO 255
        sDeviceInformation(device, i) = Inp16(regData)
    NEXT

    ' Restore the Device/Head Register.
    OUT regDeviceHead, dhrValue

END SUB

SUB InfoWin (listItems$(), device%)

    DIM win AS WindowType
    win.fTitleBar = cTrue
    win.rows = 19
    win.cols = 70

    SELECT CASE device
        CASE cPrimaryMaster
            OpenWindow win, "IdDevice - Primary Master "
        CASE cPrimarySlave
            OpenWindow win, "IdDevice - Primary Slave "
        CASE cSecondaryMaster
            OpenWindow win, "IdDevice - Secondary Master "
        CASE cSecondarySlave
            OpenWindow win, "IdDevice - Secondary Slave "
    END SELECT

    DIM cmdClose AS CommandButtonType
    cmdClose.row = 21
    cmdClose.col = 35
    cmdClose.fCancel = cTrue
    cmdClose.fDefault = cTrue

    DIM lstInfo AS ListBoxType
    lstInfo.row = 6
    lstInfo.col = 11
    lstInfo.rows = 14
    lstInfo.cols = 60
    lstInfo.fText = cTrue

    DO
        CpListBox lstInfo, listItems$()
        CpCommandButton cmdClose, "   Close   "
    LOOP UNTIL cmdClose.fClicked

    CloseWindow

END SUB

SUB MainWin

    OpenScreen cWhite, cBlue, ""

    IF WindowsIsRunning THEN
        DIM msgWin AS WindowType
        msgWin.fTitleBar = cTrue
        msgWin.fore = cHiWhite
        msgWin.back = cRed
        DIM msg$(1 TO 3)
        msg$(1) = "This program should be run under"
        msg$(2) = "MS-DOS or Windows 9x MS-DOS mode"
        junk = MessageBox(msgWin, cMbOk, 1, "IdDevice ", msg$())
        'CloseWindow
        'END
    END IF

    DIM win AS WindowType
    win.fTitleBar = cTrue
    win.rows = 12
    win.cols = 70
    OpenWindow win, "IdDevice"

    DIM cmdViewInfo AS CommandButtonType
    cmdViewInfo.row = 17
    cmdViewInfo.col = 18
    cmdViewInfo.fDefault = cTrue

    DIM cmdRescan AS CommandButtonType
    cmdRescan.row = 17
    cmdRescan.col = 36

    DIM cmdClose AS CommandButtonType
    cmdClose.row = 17
    cmdClose.col = 51
    cmdClose.fCancel = cTrue

    DIM lstDevice AS ListBoxType
    lstDevice.row = 9
    lstDevice.col = 10
    lstDevice.rows = 6
    lstDevice.cols = 62
    lstDevice.fDoubleClick = cTrue
    DIM devices$(1 TO 4)

    ' Force the initial device scan.
    cmdRescan.fClicked = cTrue

    DO
        CpListBox lstDevice, devices$()
        CpCommandButton cmdViewInfo, "  &View Info  "
        CpCommandButton cmdRescan, "  &Rescan  "
        CpCommandButton cmdClose, "   Close   "
        IF cmdRescan.fClicked THEN
            PrintStaticText 15, 33, "Scanning devices..."
            ScanDevices devices$()
            PrintStaticText 15, 33, "                   "
            lstDevice.fUpdate = cTrue
            IF lstDevice.selectedItem = 0 THEN lstDevice.selectedItem = 1
        END IF
        IF cmdViewInfo.fClicked THEN
            SELECT CASE lstDevice.selectedItem
                CASE 1
                    IF sfDeviceDetected(cPrimaryMaster) THEN
                        REDIM listItems$(0)
                        MakeList listItems$(), cPrimaryMaster
                        InfoWin listItems$(), cPrimaryMaster
                    END IF
                CASE 2
                    IF sfDeviceDetected(cPrimarySlave) THEN
                        REDIM listItems$(0)
                        MakeList listItems$(), cPrimarySlave
                        InfoWin listItems$(), cPrimarySlave
                    END IF
                CASE 3
                    IF sfDeviceDetected(cSecondaryMaster) THEN
                        REDIM listItems$(0)
                        MakeList listItems$(), cSecondaryMaster
                        InfoWin listItems$(), cSecondaryMaster
                    END IF
                CASE 4
                    IF sfDeviceDetected(cSecondarySlave) THEN
                        REDIM listItems$(0)
                        MakeList listItems$(), cSecondarySlave
                        InfoWin listItems$(), cSecondarySlave
                    END IF
            END SELECT
        END IF

    LOOP UNTIL cmdClose.fClicked

    ' Close the window opened by the call to OpenWindow.
    CloseWindow

    ' Close the window opened by the call to OpenScreen.
    CloseWindow

END SUB

SUB MakeList (listItems$(), device%)
   
    ' Reads selected portions of the device information for the device
    ' specified by <device%> from the shared device information array,
    ' formats it, and copies it to the elements of <listItems$()>.
    '
    ' The Identify Device command returns a large amount of information,
    ' most of which is intended for use by the system BIOS or by device
    ' drivers and other programs that require low-level control of the
    ' device. This procedure selects only those device information
    ' fields that contain information of general interest and/or
    ' information that may aid in the configuration and testing of
    ' a system.

    ' If bit 2 of word 0 is set, then the Identify Device
    ' response was incomplete.
    IF BitRangeValue(sDeviceInformation(device, 0), 2, 2) THEN
        AddListItem listItems$(), "Identify Device response incomplete - rescan the device"
        EXIT SUB
    END IF
    
    IF BitRangeValue(sDeviceInformation(device, 0), 15, 15) THEN
        AddListItem listItems$(), "Device Protocol : ATAPI"
        fATAPI = cTrue
    ELSE
        AddListItem listItems$(), "Device Protocol : ATA"
    END IF

    IF BitRangeValue(sDeviceInformation(device, 0), 7, 7) THEN
        AddListItem listItems$(), "Device Uses Removable Media : Yes"
    ELSE
        AddListItem listItems$(), "Device Uses Removable Media : No"
    END IF
    
    IF fATAPI THEN

        ' Return a specific device type only for the most common types.
        SELECT CASE BitRangeValue(sDeviceInformation(device, 0), 8, 12)
            CASE 0
                AddListItem listItems$(), "ATAPI Device Type : Direct-Access"
            CASE 1
                AddListItem listItems$(), "ATAPI Device Type : Sequential-Access"
            CASE 5
                AddListItem listItems$(), "ATAPI Device Type : CD-ROM"
            CASE ELSE
                AddListItem listItems$(), "ATAPI Device Type : ?"
        END SELECT
    END IF

    IF NOT fATAPI THEN
        AddListItem listItems$(), "Default Logical Cylinders : " + GroupedDigits$(STR$(sDeviceInformation(device, 1)))
        AddListItem listItems$(), "Default Logical Heads : " + GroupedDigits$(STR$(sDeviceInformation(device, 3)))
        AddListItem listItems$(), "Default Logical Sectors Per Track : " + GroupedDigits$(STR$(sDeviceInformation(device, 6)))
    END IF

    AddListItem listItems$(), "Serial Number : " + SerialNumber$(device)
    AddListItem listItems$(), "Firmware Revision : " + FirmwareRevision$(device)
    AddListItem listItems$(), "Model Number : " + ModelNumber$(device)

    IF NOT fATAPI THEN
        AddListItem listItems$(), "Maximum Sectors Per Block For Read/Write Multiple : " + GroupedDigits$(STR$(BitRangeValue(sDeviceInformation(device, 47), 0, 7)))
    END IF

    IF BitRangeValue(sDeviceInformation(device, 49), 11, 11) THEN
        AddListItem listItems$(), "IORDY Flow Control Support : Yes"
    ELSE
        AddListItem listItems$(), "IORDY Flow Control Support : ?"
    END IF

    ' In the ATA/ATAPI-5 standard, for ATA devices only, bit 8 and 9
    ' are fixed at 1 for backwards compatibility.
    IF BitRangeValue(sDeviceInformation(device, 49), 9, 9) THEN
        AddListItem listItems$(), "LBA Support : Yes"
    ELSE
        AddListItem listItems$(), "LBA Support : No"
    END IF
    IF BitRangeValue(sDeviceInformation(device, 49), 8, 8) THEN
        AddListItem listItems$(), "DMA Support : Yes"
    ELSE
        AddListItem listItems$(), "DMA Support : No"
    END IF

    IF BitRangeValue(sDeviceInformation(device, 53), 0, 0) AND (NOT fATAPI) THEN
        AddListItem listItems$(), "Current Logical Cylinders : " + GroupedDigits$(STR$(sDeviceInformation(device, 54)))
        AddListItem listItems$(), "Current Logical Heads : " + GroupedDigits$(STR$(sDeviceInformation(device, 55)))
        AddListItem listItems$(), "Current Logical Sectors Per Track : " + GroupedDigits$(STR$(sDeviceInformation(device, 56)))
        AddListItem listItems$(), "Current Capacity In Sectors : " + GroupedDigits$(STR$(&H10000 * UInt(sDeviceInformation(device, 58)) + UInt(sDeviceInformation(device, 57))))
    END IF
   
    IF BitRangeValue(sDeviceInformation(device, 59), 8, 8) AND (NOT fATAPI) THEN
        AddListItem listItems$(), "Current Sectors Per Block For Read/Write Multiple : " + GroupedDigits$(STR$(BitRangeValue(sDeviceInformation(device, 59), 0, 7)))
    END IF

    IF BitRangeValue(sDeviceInformation(device, 49), 9, 9) AND (NOT fATAPI) THEN
        
        totalSectorsLBA& = &H10000 * UInt(sDeviceInformation(device, 61)) + UInt(sDeviceInformation(device, 60))
        AddListItem listItems$(), "Total Sectors Addressable In LBA Mode : " + GroupedDigits$(STR$(totalSectorsLBA&))
        AddListItem listItems$(), "Total Bytes In LBA Mode (calculated) : " + GroupedDigits$(STR$(512# * totalSectorsLBA&))
        
        currentHeads& = sDeviceInformation(device, 55)
        currentSecPerTrack& = sDeviceInformation(device, 56)
        IF currentHeads& AND currentSecPerTrack& THEN
            GetChs cylinders&, h&, s&, totalSectorsLBA&, currentHeads&, currentSecPerTrack&
            AddListItem listItems$(), "Cylinders In LBA Mode (calculated) : " + GroupedDigits$(STR$(cylinders&))
        END IF
    END IF

    IF BitRangeValue(sDeviceInformation(device, 49), 8, 8) THEN
        temp$ = ""

        ' Bits 3-7 and 11-15 are reserved in the ATA/ATAPI-5 standard.
        FOR i = 0 TO 2
            IF BitRangeValue(sDeviceInformation(device, 63), i, i) THEN
                temp$ = temp$ + RTRIM$(STR$(i)) + ","
            END IF
        NEXT
        IF RIGHT$(temp$, 1) = "," THEN temp$ = LEFT$(temp$, LEN(temp$) - 1)
        AddListItem listItems$(), "Multiword DMA Transfer Modes Supported :" + temp$
        FOR i = 8 TO 10
            IF BitRangeValue(sDeviceInformation(device, 63), i, i) THEN
                AddListItem listItems$(), "Current Multiword DMA Transfer Mode :" + STR$(i - 8)
                EXIT FOR
            END IF
        NEXT
    END IF
    
    IF BitRangeValue(sDeviceInformation(device, 53), 1, 1) THEN
        temp$ = ""
        FOR i = 0 TO 7
            IF BitRangeValue(sDeviceInformation(device, 64), i, i) THEN
                temp$ = temp$ + RTRIM$(STR$(i + 3)) + ","
            END IF
        NEXT
        IF RIGHT$(temp$, 1) = "," THEN temp$ = LEFT$(temp$, LEN(temp$) - 1)
        IF temp$ <> "" THEN
            AddListItem listItems$(), "Flow Control PIO Modes Supported :" + temp$
        END IF

        IF sDeviceInformation(device, 67) THEN
            AddListItem listItems$(), "Minimum PIO Cycle Time Without Flow Control (ns) :" + STR$(sDeviceInformation(device, 67))
        ELSE
            AddListItem listItems$(), "Minimum PIO Cycle Time Without Flow Control (ns) : NA"
        END IF

        IF sDeviceInformation(device, 68) THEN
            AddListItem listItems$(), "Minimum PIO Cycle Time With Flow Control (ns) :" + STR$(sDeviceInformation(device, 68))
        ELSE
            AddListItem listItems$(), "Minimum PIO Cycle Time With Flow Control (ns) : NA"
        END IF
    END IF

    ' In the ATA/ATAPI-5 standard, bits 6-14 of word 80 are reserved.
    ' This code tests the reserved bits on the assumption that
    ' compliance with newer versions of the ATA/ATAPI standard will
    ' be indicated in the same manner as for the current versions.
    IF sDeviceInformation(device, 80) <> 0 AND sDeviceInformation(device, 80) <> &HFFFF THEN
        FOR i = 14 TO 2 STEP -1
            IF BitRangeValue(sDeviceInformation(device, 80), i, i) THEN
                IF i = 2 OR i = 3 THEN
                    standard$ = "ATA-"
                ELSE
                    standard$ = "ATA/ATAPI-"
                END IF
                standard$ = standard$ + LTRIM$(STR$(i))
                AddListItem listItems$(), "Device claims compliance with : " + standard$
                EXIT FOR
            END IF
        NEXT
    END IF

    ' In the ATA/ATAPI-5 standard, bits 5-7 and 13-15 of word 88 are
    ' reserved. This code tests the reserved bits on the assumption
    ' that any new Ultra DMA modes will be indicated in the same
    ' manner as the current modes.
    IF BitRangeValue(sDeviceInformation(device, 53), 2, 2) THEN
        FOR i = 7 TO 0 STEP -1
            IF BitRangeValue(sDeviceInformation(device, 88), i, i) THEN
                AddListItem listItems$(), "Maximum Ultra DMA mode supported :" + STR$(i)
                maxUdmaModeSupported = i
                EXIT FOR
            END IF
        NEXT
        FOR i = 8 TO 15
            IF BitRangeValue(sDeviceInformation(device, 88), i, i) THEN
                AddListItem listItems$(), "Current Ultra DMA mode :" + STR$(i - 8)
                EXIT FOR
            END IF
        NEXT
    END IF
    
    ' In the ATA/ATAPI-5 standard, bit 13 of word 93 reports the result
    ' of the devices's sampling of PDIAG-:CBLID- (Passed diagnostics:
    ' Cable assembly type identifier). The host system uses this value,
    ' in combination with other information, to determine the cable
    ' type (40 or 80 conductor). Considered alone, this value is
    ' meaningful only under specific circumstances:
    '   IF the host and connected devices (properly) support Ultra DMA
    '   modes greater than 2 AND an 80-conductor cable is used to
    '   connect the devices to the host AND the correct end of the
    '   cable is connected to the host, THEN the device *should have*
    '   sampled CBLID- above Vih (Voltage input high).
    IF maxUdmaModeSupported > 2 THEN
        IF BitRangeValue(sDeviceInformation(device, 93), 13, 13) THEN
            AddListItem listItems$(), "Device sampled CBLID- above Vih"
        ELSE
            AddListItem listItems$(), "Device sampled CBLID- below Vih"
        END IF
    END IF

END SUB

FUNCTION ModelNumber$ (device%)
   
    ' Returns the 40 character model number for the device specified by
    ' <device%> from the shared device information array.

    ' Return a null string if the model number is not specified.
    IF sDeviceInformation(device, 27) = 0 THEN EXIT FUNCTION
   
    FOR i = 27 TO 46
        r$ = r$ + CHR$(HighByte(sDeviceInformation(device, i))) + CHR$(LowByte(sDeviceInformation(device, i)))
    NEXT
    ModelNumber$ = LTRIM$(r$)

END FUNCTION

SUB ScanDevices (devices$())

    ' Checks the standard ATA interface channels for devices that
    ' support the ATA or ATAPI Identify Device command. For each
    ' device found, sets the appropriate element of the shared
    ' detection state array, reads the device information and
    ' copies it to the shared device information array, and loads
    ' the name of the channel/device and a model number or an
    ' error message into each element of <devices$()>.
    '
    ' This procedure issues the Identify Device command to the slave
    ' device first, as recommended in the ATA/ATAPI standard.

    DIM errorCode(1 TO 4)

    IF LBOUND(devices$) <> 1 THEN ERROR cErrorIllegalFunctionCall
    IF UBOUND(devices$) < 4 THEN ERROR cErrorIllegalFunctionCall

    ' Clear the shared flags in case this is not the first call.
    sfDeviceDetected(cPrimaryMaster) = cFalse
    sfDeviceDetected(cPrimarySlave) = cFalse
    sfDeviceDetected(cSecondaryMaster) = cFalse
    sfDeviceDetected(cSecondarySlave) = cFalse

    IdentifyDevice cPrimarySlave, errorCode(cPrimarySlave)
    Delay 500
    IdentifyDevice cPrimaryMaster, errorCode(cPrimaryMaster)
    Delay 500
    IdentifyDevice cSecondarySlave, errorCode(cSecondarySlave)
    Delay 500
    IdentifyDevice cSecondaryMaster, errorCode(cSecondaryMaster)

    temp$ = "Primary Slave   : "
    SELECT CASE errorCode(cPrimarySlave)
        CASE 0
            temp$ = temp$ + ModelNumber(cPrimarySlave)
            sfDeviceDetected(cPrimarySlave) = cTrue
        CASE cIdErrorNoDevice
            ' This error value normally indicates that
            ' there are no devices connected to the ATA
            ' interface channel (or that the program is
            ' running under a version of Windows 9x
            ' that blocks access to the Command/Control
            ' Block registers).
            temp$ = temp$ + "Device not detected"
        CASE cIdErrorTimeoutBSY
            temp$ = temp$ + "Timeout device busy"
        CASE cIdErrorTimeoutNoDRQ
            ' If the other device connected to the
            ' channel is present, then this error value
            ' normally indicates that the target device
            ' is not present.
            IF errorCode(cPrimaryMaster) = 0 THEN
                temp$ = temp$ + "Device not detected"
            ELSE
                temp$ = temp$ + "* Unexpected error *"
            END IF
        CASE cIdErrorUnexpected
            temp$ = temp$ + "* Unexpected error *"
    END SELECT
    devices$(cPrimarySlave) = temp$

    temp$ = "Primary Master  : "
    SELECT CASE errorCode(cPrimaryMaster)
        CASE 0
            temp$ = temp$ + ModelNumber(cPrimaryMaster)
            sfDeviceDetected(cPrimaryMaster) = cTrue
        CASE cIdErrorNoDevice
            temp$ = temp$ + "Device not detected"
        CASE cIdErrorTimeoutBSY
            temp$ = temp$ + "Timeout device busy"
        CASE cIdErrorTimeoutNoDRQ
            IF errorCode(cPrimarySlave) = 0 THEN
                temp$ = temp$ + "Device not detected"
            ELSE
                temp$ = temp$ + "* Unexpected error *"
            END IF
        CASE cIdErrorUnexpected
            temp$ = temp$ + "* Unexpected error *"
    END SELECT
    devices$(cPrimaryMaster) = temp$

    temp$ = "Secondary Slave : "
    SELECT CASE errorCode(cSecondarySlave)
        CASE 0
            temp$ = temp$ + ModelNumber(cSecondarySlave)
            sfDeviceDetected(cSecondarySlave) = cTrue
        CASE cIdErrorNoDevice
            temp$ = temp$ + "Device not detected"
        CASE cIdErrorTimeoutBSY
            temp$ = temp$ + "Timeout device busy"
        CASE cIdErrorTimeoutNoDRQ
            IF errorCode(cSecondaryMaster) = 0 THEN
                temp$ = temp$ + "Device not detected"
            ELSE
                temp$ = temp$ + "* Unexpected error *"
            END IF
        CASE cIdErrorUnexpected
            temp$ = temp$ + "* Unexpected error *"
    END SELECT
    devices$(cSecondarySlave) = temp$

    temp$ = "Secondary Master: "
    SELECT CASE errorCode(cSecondaryMaster)
        CASE 0
            temp$ = temp$ + ModelNumber(cSecondaryMaster)
            sfDeviceDetected(cSecondaryMaster) = cTrue
        CASE cIdErrorNoDevice
            temp$ = temp$ + "Device not detected"
        CASE cIdErrorTimeoutBSY
            temp$ = temp$ + "Timeout device busy"
        CASE cIdErrorTimeoutNoDRQ
            IF errorCode(cSecondarySlave) = 0 THEN
                temp$ = temp$ + "Device not detected"
            ELSE
                temp$ = temp$ + "* Unexpected error *"
            END IF
        CASE cIdErrorUnexpected
            temp$ = temp$ + "* Unexpected error *"
    END SELECT
    devices$(cSecondaryMaster) = temp$

END SUB

FUNCTION SerialNumber$ (device%)

    ' Returns the 20 character serial number for the device specified
    ' by <device%> from the shared device information array.
   
    ' Return a null string if the serial number is not specified.
    IF sDeviceInformation(device, 10) = 0 THEN EXIT FUNCTION
   
    FOR i = 10 TO 19
        r$ = r$ + CHR$(HighByte(sDeviceInformation(device, i))) + CHR$(LowByte(sDeviceInformation(device, i)))
    NEXT
    SerialNumber$ = LTRIM$(r$)

END FUNCTION

